from __future__ import annotations
from dataclasses import dataclass, field
import os
from pathlib import Path

import pytest

from eyconf import EYConf


@pytest.fixture(autouse=True)
def tmp_config_path(tmp_path) -> Path:
    config_file_path = tmp_path / "config.yml"
    os.environ["EYCONF_CONFIG_FILE"] = str(config_file_path)
    return config_file_path


@dataclass
class Config42:
    int_field: int = 42
    str_field: str = field(default="FortyTwo!")


class TestEYConfFile:
    """Tests the default EYConf configuration file handling."""

    def test_init(self, tmp_config_path):
        conf = EYConf(Config42)
        assert isinstance(conf, EYConf)
        assert conf._schema == Config42

        with pytest.raises(ValueError):
            EYConf("not a dataclass")  # type: ignore

    def test_init_bad_defaults(self):
        # Schema with bad defaults
        with pytest.raises(TypeError):

            @dataclass
            class BadConfig:
                foo: Config42

            EYConf(BadConfig)

    def test_write_default(self, tmp_config_path, caplog):
        conf = EYConf(Config42)

        assert conf.path == tmp_config_path
        assert conf.path.exists()
        assert conf.path.read_text() == "int_field: 42\nstr_field: FortyTwo!\n"

        # Overwrite again
        with caplog.at_level("WARNING"):
            conf._write_default()
            assert "overwriting" in caplog.text.lower()

    def test_load_existing(self, tmp_config_path):
        with open(tmp_config_path, "w") as f:
            f.write("int_field: 20\nstr_field: Another value!\n")
        conf = EYConf(Config42)
        assert conf._data.int_field == 20
        assert conf._data.str_field == "Another value!"

    def test_reload(self, tmp_config_path):
        with open(tmp_config_path, "w") as f:
            f.write("int_field: 10\nstr_field: Temp value!\n")
        conf = EYConf(Config42)
        assert conf._data.int_field == 10
        assert conf._data.str_field == "Temp value!"

        # Modify the config file
        with open(tmp_config_path, "w") as f:
            f.write("int_field: 99\nstr_field: Changed value!\n")

        conf.reload()
        assert conf._data.int_field == 99
        assert conf._data.str_field == "Changed value!"

        # Should raise for missing config
        tmp_config_path.unlink()
        with pytest.raises(FileNotFoundError):
            conf.reload()

    def test_reset(self, tmp_config_path):
        with open(tmp_config_path, "w") as f:
            f.write("int_field: 5\nstr_field: Old value!\n")
        conf = EYConf(Config42)
        assert conf._data.int_field == 5
        assert conf._data.str_field == "Old value!"

        conf.reset()
        assert conf._data.int_field == 42
        assert conf._data.str_field == "FortyTwo!"

    def test_repr(self, tmp_config_path):
        conf = EYConf(Config42)
        repr_str = repr(conf)
        assert "EYConf" in repr_str
        assert "int_field" in repr_str
        assert "str_field" in repr_str

    def test_default_validate(self, tmp_config_path):
        with open(tmp_config_path, "w") as f:
            f.write("int_field: 5\n")  # omit str_field to use default
        conf = EYConf(Config42)
        conf._data.str_field = "FortyTwo!"



from dataclasses import dataclass, field

import pytest
from eyconf.decorators import (
    DictSetAccess,
    allow_additional,
    check_allows_additional,
    dict_access,
    DictAccess,
)


@dataclass
class SchemaUndercorated:
    foo: int = 1


@dict_access(
    getter=True,
    setter=True,
)
@dataclass
class SchemaWithDictAcecess:
    foo: int = 1


@allow_additional
@dataclass
class SchemaWithAllowAdditional:
    foo: int = 1


@allow_additional
@dict_access
@dataclass
class SchemaWithBoth:
    foo: int = 1


class TestAllowsAdditional:
    def test_allow_additional_sets_flag(self):
        config1 = SchemaWithAllowAdditional()
        assert check_allows_additional(SchemaWithAllowAdditional)
        assert check_allows_additional(config1)

        config2 = SchemaWithBoth()
        assert check_allows_additional(SchemaWithBoth)
        assert check_allows_additional(config2)

    def test_no_allow_additional_flag(self):
        config = SchemaUndercorated()
        assert not check_allows_additional(SchemaUndercorated)
        assert not check_allows_additional(config)


class TestDictAccess:
    def test_item_assignment(self):
        """Test basic dict-style get access."""
        obj = SchemaWithDictAcecess()
        assert isinstance(obj, DictAccess)
        assert isinstance(obj, DictSetAccess)
        obj["foo"] = 10
        assert obj["foo"] == 10

    def test_attribute_assignment(self):
        """Test basic dict-style set access."""
        obj = SchemaWithDictAcecess()
        obj.foo = 20
        assert obj.foo == 20


class TestDictAccessWithAlias:
    @pytest.fixture(autouse=True)
    def alias_fixture(self):
        @dict_access(getter=True, setter=True)
        @dataclass
        class AliasConfig:
            attr_field: int = field(metadata={"alias": "dict_field"})
            str_field: str = "FortyTwo!"

        self.AliasConfig = AliasConfig

    def test_item_assignment_with_alias(self):
        """Test dict-style get/set access with alias resolution."""
        obj = self.AliasConfig(attr_field=42)
        assert isinstance(obj, DictAccess)
        assert isinstance(obj, DictSetAccess)

        # Access via alias
        assert obj["dict_field"] == 42

        # Set via alias
        obj["dict_field"] = 100
        assert obj["dict_field"] == 100

        # Access via attribute name should raise
        with pytest.raises(KeyError):
            _ = obj["attr_field"]

        with pytest.raises(KeyError):
            obj["attr_field"] = 150

    def test_attribute_assignment_with_alias(self):
        """Test attribute get/set access with alias resolution."""
        obj = self.AliasConfig(attr_field=42)

        # Access via attribute name ONLY
        assert obj.attr_field == 42

        # Set via attribute name ONLY
        obj.attr_field = 200
        assert obj.attr_field == 200

        with pytest.raises(AttributeError):
            _ = obj.dict_field  # type: ignore



"""Generate a default yaml file from a dataclass schema.

This is relatively dirty and not all edge cases might be covered. Works for most
simple cases though.
"""

from __future__ import annotations

import logging
import re
from abc import ABC, abstractmethod
from dataclasses import _MISSING_TYPE, Field, fields, is_dataclass
from types import NoneType, UnionType
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    get_args,
    get_origin,
)

from eyconf.constants import primitive_types
from eyconf.type_utils import get_type_hints_resolve_namespace

log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

    from eyconf.constants import Primitives


class Line(ABC):
    """A line of yaml content."""

    is_comment: bool = False
    indent: int = 0

    def __init__(self, is_comment: bool = False, indent: int = 0):
        self.is_comment = is_comment
        self.indent = indent

    @property
    @abstractmethod
    def _content(self) -> str:
        """Return the formatted line. Without indentation or comment."""
        pass

    @property
    def content(self) -> str:
        """Return the formatted line."""
        return f"{'  ' * self.indent}{'# ' if self.is_comment else ''}{self._content}"


class EmptyLine(Line):
    """An empty line of yaml content."""

    def __init__(self):
        super().__init__(is_comment=False, indent=0)

    @property
    def _content(self) -> str:
        """Return the formatted line. Without indentation or comment."""
        return ""

    @property
    def content(self) -> str:
        """Return the formatted line."""
        return ""


class CommentLine(Line):
    """A comment line of yaml content."""

    def __init__(self, comment: str, indent: int = 0):
        super().__init__(is_comment=True, indent=indent)
        self.comment = comment

    @property
    def _content(self) -> str:
        """Return the formatted line. Without indentation or comment."""
        return self.comment.strip()


class MapLine(Line):
    """A map line of yaml content."""

    name: str
    default_value: Primitives

    def __init__(self, name: str, default_value: Primitives, **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.default_value = default_value

    @property
    def _content(self) -> str:
        """Return the formatted line. Without indentation or comment."""
        value = self.default_value
        if value is None:
            value = "null"
        elif isinstance(value, bool):
            value = str(value).lower()
        return f"{self.name}: {value}"


class SequenceLine(Line):
    """A sequence line of yaml content."""

    default_value: Primitives

    def __init__(self, default_value: Primitives, **kwargs):
        super().__init__(**kwargs)
        self.default_value = default_value

    @property
    def _content(self) -> str:
        """Return the formatted line. Without indentation or comment."""
        return f"- {self.default_value}"


class SectionLine(Line):
    """A section line of yaml content."""

    name: str

    def __init__(self, name: str, **kwargs):
        super().__init__(**kwargs)
        self.name = name

    @property
    def _content(self) -> str:
        """Return the formatted line. Without indentation or comment."""
        return f"{self.name}:"


def dataclass_to_yaml(schema: type[DataclassInstance] | DataclassInstance) -> str:
    """Generate a yaml string from a dataclass schema."""
    lines = _dataclass_to_lines(schema)
    return "\n".join([line.content for line in lines])


def _dataclass_to_lines(
    schema: type[DataclassInstance] | DataclassInstance,
    indent: int = 0,
) -> list[Line]:
    """Generate yaml from a dataclass schema."""
    lines: list[Line] = []

    # Parse docstring
    if schema.__doc__ is not None and __is_custom_docstring(schema):
        lines += __split_docstring(schema.__doc__, indent=indent)
        lines.append(EmptyLine())

    # Handle dataclass instances
    if __is_dataclass_instance(schema):
        dict = schema.__dict__
        for key, value in dict.items():
            lines.append(MapLine(name=key, default_value=value, indent=indent))
        return lines

    # Handle dataclass types
    # by parsing type hint
    dataclass_types = get_type_hints_resolve_namespace(
        schema,
        include_extras=True,
    )
    all_fields = fields(schema)

    for field in all_fields:
        lines += __field_to_lines(field, dataclass_types[field.name], indent=indent)

    return lines


def __field_to_lines(field: Field[Any], field_type: type, indent=0) -> list[Line]:
    """Parse a primitive field and return a list of lines.

    Parameters
    ----------
    field : Field
        The field to parse.
    field_type : type
        The type of the field. Parsed here additionally
        to the included type in the field to support
        `from __future__ import annotations`. This is basically
        an overwrite of `field.type`.

    Returns
    -------
    list[str]
        A list of lines, each no longer than `l` characters.
    """
    lines: list[Line] = []
    origin = get_origin(field_type)
    args = get_args(field_type)

    # Extract docstring from annotated
    if origin is Annotated:
        # We always assume the first argument is the type
        # And all following are annotations (must be strings)
        field_type = args[0]
        annotations = [arg for arg in args[1:] if isinstance(arg, str)]
        for annotation in annotations:
            lines += __split_docstring(annotation, indent=indent)

    if is_dataclass(field_type):
        # Add section
        lines.append(SectionLine(field.name, indent=indent))
        lines += _dataclass_to_lines(field_type, indent=indent + 1)
        lines.append(EmptyLine())
        return lines

    # Check if field is optional
    is_optional = False
    if origin is UnionType:
        is_optional = type(None) in args or NoneType in args
        args = tuple(
            arg for arg in args if arg is not type(None) and arg is not NoneType
        )

    # Parse default
    default_missing = True
    default_value: Any = None
    if isinstance(field.default, _MISSING_TYPE):
        if isinstance(field.default_factory, _MISSING_TYPE):
            default_missing = True
        else:
            default_value = field.default_factory()
            default_missing = False
    else:
        default_value = field.default
        default_missing = False

    if origin in [dict, dict] and default_missing:
        default_value = {}
        default_missing = False

    # No default value only allowed if the field is optional
    if default_missing and not is_optional:
        raise ValueError(
            f"Field '{field.name}' has no default value! You may set one using direct assignment or a default factory."
        )

    # Default value: Lists/Sequences
    if isinstance(default_value, list):
        if len(default_value) == 0:
            default_value = None
        else:
            """
            Stuff:
                - 1
                - 2
                - 3
            """
            lines.append(SectionLine(field.name, indent=indent))
            for value in default_value:
                if __is_primitive_instance(value):
                    lines.append(SequenceLine(default_value=value, indent=indent + 1))
                    continue

                if __is_dataclass_instance(value):
                    lines.append(SequenceLine("", indent=indent + 1))
                    lines += _dataclass_to_lines(value, indent=indent + 2)
                    continue

                raise NotImplementedError(
                    f"Field type {field.type} {args} {origin} is not supported."
                )

            return lines

    # Default value: dicts
    if isinstance(default_value, dict):
        if len(default_value) == 0:
            """
            Stuff: {}
            """
            lines.append(MapLine(name=field.name, default_value=r"{}", indent=indent))
            return lines
        else:
            """
            Stuff:
                key1: value1
                key2: value2
            """
            lines.append(SectionLine(field.name, indent=indent))
            for key, value in default_value.items():
                if not isinstance(key, str):
                    raise ValueError("Only string keys are supported in dict types")

                if __is_primitive_instance(value):
                    lines.append(
                        MapLine(name=key, default_value=value, indent=indent + 1)
                    )
                    continue

                if __is_dataclass_instance(value):
                    lines.append(MapLine(name=key, default_value="", indent=indent + 1))
                    lines += _dataclass_to_lines(value, indent=indent + 2)
                    continue

                raise NotImplementedError(
                    f"Field type {field.type} {args} {origin} is not supported."
                )

            return lines

    # Might not type check but it is a good enough heuristic
    # if default_value is set wrongly there are more issues anyways
    if __is_primitive_instance(default_value) or is_optional:
        lines.append(
            MapLine(name=field.name, default_value=default_value, indent=indent)
        )
        return lines

    raise NotImplementedError(
        f"Field type {field.type} {args} {origin} is not supported."
    )


def __split_docstring(docstring: str, l=80, indent=0) -> list[Line]:
    """Parse a docstring and return a list of lines.

    Tries to split full words if possible.

    Parameters
    ----------
    docstring : str
        The docstring to parse.

    l : int
        The line length.

    Returns
    -------
    list[str]
        A list of lines, each no longer than `l` characters.
    """
    lines = []
    for line in docstring.split("\n"):
        while len(line) > l:
            split_at = line.rfind(" ", 0, l)
            if split_at == -1:
                # If no space is found, split at the maximum length
                split_at = l
            lines.append(line[:split_at])
            line = line[split_at:].lstrip()
        lines.append(line)
    return [CommentLine(line.strip(), indent=indent) for line in lines if line.strip()]


def __is_custom_docstring(dataclass_obj):
    # Create a regex pattern to match the default docstring format
    default_docstring_pattern = r"^[^\(]+\([^\)]*\)(\w*|.*)$"

    # Check if the docstring matches the default pattern
    return re.match(default_docstring_pattern, dataclass_obj.__doc__) is None


def __is_primitive_instance(t: Any) -> bool:
    """Check if the field type is a primitive instance.

    Parameters
    ----------
    field : Field
        The field to check.

    Returns
    -------
    bool
        True if the field type is a primitive instance, False otherwise.
    """
    if isinstance(t, (*primitive_types,)):
        return True
    return False


def __is_dataclass_instance(t: Any) -> bool:
    """Check if the field type is a dataclass instance.

    Parameters
    ----------
    field : Field
        The field to check.

    Returns
    -------
    bool
        True if the field type is a dataclass instance, False otherwise.
    """
    return is_dataclass(t) and not isinstance(t, type)


__all__ = [
    "dataclass_to_yaml",
]



from __future__ import annotations

import logging
from collections.abc import Iterable
from dataclasses import Field, fields, is_dataclass
from types import NoneType, UnionType
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    TypedDict,
    TypeVar,
    get_args,
    get_origin,
    get_type_hints,
)

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

log = logging.getLogger(__name__)

D = TypeVar("D", bound="DataclassInstance")
T = TypeVar("T")


def merge_dicts(
    a: dict, b: dict, path=[], priority: Literal["raise", "a", "b"] = "raise"
) -> dict:
    """Merge dict b into dict a, raising an exception on conflicts."""
    # Convert path to tuple for faster concatenation
    path_tuple = tuple(path)

    for key in b:
        val_b = b[key]
        if key in a:
            val_a = a[key]
            # Handle nested dictionaries recursively
            if isinstance(val_a, dict) and isinstance(val_b, dict):
                merge_dicts(val_a, val_b, path_tuple + (str(key),), priority)
            elif val_a != val_b:
                # Handle conflicts based on priority
                if priority == "a":
                    # Keep a's value, ignore b's value
                    continue
                elif priority == "b":
                    # Use b's value, overwriting a's value
                    a[key] = val_b
                else:
                    full_path = ".".join(path_tuple + (str(key),))
                    raise Exception(f"Conflict at {full_path}: {val_a} != {val_b}")
        else:
            a[key] = val_b

    return a


def dataclass_from_dict(in_type: type[D], data: dict) -> D:
    """Convert a dict to a dataclass instance of the given type. Always returns a dataclass."""
    result = _dataclass_from_dict_inner(in_type, data)
    if result is None:
        raise ValueError(f"Could not parse data {data} with type {in_type}")
    return result


def _dataclass_from_dict_inner(target_type: type, data: Any) -> Any:
    """Inner function that handles Union types and may return None."""
    # Handle Union types
    origin = get_origin(target_type)
    if origin is UnionType:
        args = get_args(target_type)
        includes_none = any(arg is NoneType or arg is type(None) for arg in args)

        if data is None and includes_none:
            return None

        for arg in args:
            if arg is NoneType or arg is type(None):
                continue
            try:
                return _dataclass_from_dict_inner(arg, data)
            except (ValueError, TypeError, KeyError):
                continue
        return None

    # Handle dict data - convert to dataclass
    if isinstance(data, dict) and is_dataclass(target_type):
        field_types = get_type_hints(target_type, include_extras=False)
        found_fields = {}
        additional_fields = {}

        aliased_fields = {}  # alias to name, only aliased fields
        field_types_to_use = {}  # name to type, all fields
        for f in fields(target_type):
            if hasattr(f, "metadata"):
                aliased_fields[f.metadata.get("alias")] = f.name
            field_types_to_use[f.name] = field_types.get(f.name, f.type)

        for key, value in data.items():
            if key in aliased_fields.keys():
                key = aliased_fields[key]
                found_fields[key] = _dataclass_from_dict_inner(
                    field_types_to_use[key], value
                )
            elif key in field_types_to_use.keys():
                found_fields[key] = _dataclass_from_dict_inner(
                    field_types_to_use[key], value
                )
            else:
                additional_fields[key] = value

        try:
            res = target_type(**found_fields)  # type: ignore[bad-instantiation]
            if len(additional_fields) > 0:
                log.warning(
                    f"Additional fields {list(additional_fields.keys())} "
                    f"found for dataclass {target_type.__name__} but "
                    "not in schema and will be ignored."
                )
            return res
        except TypeError as e:
            raise ValueError(f"Failed to create {target_type.__name__}: {e}")

    # Potentially nested dataclass in dicts
    if isinstance(data, dict) and get_origin(target_type) is dict:
        key_type, value_type = get_args(target_type)
        return {
            _dataclass_from_dict_inner(key_type, k): _dataclass_from_dict_inner(
                value_type, v
            )
            for k, v in data.items()
        }

    # Handle sequence types (list, tuple)
    if isinstance(data, (list, tuple)):
        if hasattr(target_type, "__args__") and target_type.__args__:
            elem_type = target_type.__args__[0]
            return [_dataclass_from_dict_inner(elem_type, item) for item in data]
        else:
            return data

    # Handle primitive types
    return data


class Metadata(TypedDict, total=False):
    """Metadata for a dataclass field.

    This defines all metadata keys we use in eyconf for
    config schema fields.
    """

    alias: str
    """Alias for the field when used in dict representations."""


def get_metadata(
    type: type | DataclassInstance,
) -> Iterable[tuple[Field[Any], Metadata]]:
    """Extract metadata from dataclass fields."""
    dataclass_fields = fields(type) if is_dataclass(type) else {}
    return ((f, Metadata(**f.metadata)) for f in dataclass_fields if f.metadata)


def metadata_fields_from_dataclass(
    type: type | DataclassInstance,
) -> dict[str, Metadata]:
    """Extract metadata from dataclass fields."""
    return {f.name: m for f, m in get_metadata(type)}


def dict_items_resolve_aliases(
    data: dict[str, T],
    type: type | DataclassInstance,
) -> Iterable[tuple[str, T]]:
    """`dict.items()` but resolves alias names.

    Allows to iter a dictionary using the attribute style
    access keys.

    This resolve to attribute style keys (alias->non-alias).
    """
    dict_key_to_attr_key = {
        m["alias"]: f.name for f, m in get_metadata(type) if "alias" in m
    }
    return ((dict_key_to_attr_key.get(key, key), value) for key, value in data.items())



from .config import Config, ConfigExtra, EYConf
from .generate_yaml import dataclass_to_yaml
from .validation import validate

__all__ = [
    "dataclass_to_yaml",
    "EYConf",
    "Config",
    "ConfigExtra",
    "validate",
]



from collections.abc import Sequence
from typing import Annotated
import pytest
from eyconf.decorators import DictAccess, dict_access
from eyconf.type_utils import (
    iter_dataclass_type,
)
from dataclasses import dataclass, field

from eyconf.asdict import asdict_with_aliases

# for some reason typing  Sequence and abc sequence are not the same type
from typing import Sequence as TypingSequence  # noqa: UP035

from eyconf.utils import merge_dicts  # noqa: UP035


@dict_access
@dataclass
class Nested:
    str_field: str = "FortyTwo"


@dict_access
@dataclass
class Config42:
    int_field: int = 42
    nested: Nested = field(default_factory=Nested)


class TestDictStyleAccess:
    """Test the @dict_access decorator functionality."""

    def test_get_data(self):
        config = Config42()

        # Note that placing `assert isinstance(config, DictAccess)` here
        # will make some runtime type checkers error about the attribute access
        # Once you place the assert, better stick to dict-style access
        # for the rest of the scope
        assert isinstance(config, DictAccess)
        assert isinstance(config["nested"], DictAccess)

        assert config.int_field == 42
        assert config.nested.str_field == "FortyTwo"

        assert config["int_field"] == 42
        assert config["nested"]["str_field"] == "FortyTwo"


@dataclass
class ConfigWithProperty:
    int_field: int = 42

    @property
    def computed_property(self) -> str:
        return f"int_filed={self.int_field}"


class TestAsDict:
    """Test our customized asdict_with_aliases function.

    Besides the alias resolution, it also supports extracting
    properties and non-field attributes.

    Alias resolution is tested in test_alias.py
    """

    @pytest.mark.parametrize(
        "include_properties, include_attributes, expected_keys",
        [
            (False, False, {"int_field"}),
            (True, False, {"int_field", "computed_property"}),
            (False, True, {"int_field", "non_field_attr"}),
            (True, True, {"int_field", "computed_property", "non_field_attr"}),
        ],
    )
    def test_asdict(self, include_properties, include_attributes, expected_keys):
        config = ConfigWithProperty()
        config.non_field_attr = 100  # type: ignore
        dump = asdict_with_aliases(
            config,
            include_properties=include_properties,
            include_attributes=include_attributes,
        )
        assert set(dump.keys()) == expected_keys


@dataclass
class Item:
    id: int


class TestIterDataclassType:
    def test_iter_dataclass_type_basic(self):
        """Test basic dataclass iteration."""

        @dataclass
        class Inner:
            value: int

        @dataclass
        class Outer:
            inner: Inner
            name: str

        result = list(iter_dataclass_type(Outer))
        assert len(result) == 2
        assert Outer in result
        assert Inner in result

    def test_iter_dataclass_type_nested(self):
        """Test deeply nested dataclass structures."""

        @dataclass
        class Level3:
            value: int

        @dataclass
        class Level2:
            level3: Level3

        @dataclass
        class Level1:
            level2: Level2

        result = list(iter_dataclass_type(Level1))
        assert len(result) == 3
        assert Level1 in result
        assert Level2 in result
        assert Level3 in result

    @pytest.mark.parametrize(
        "collection_type",
        [
            list[Item],
            tuple[Item, Item],
            set[Item],
            Sequence[Item],
            TypingSequence[Item],
            Item | None,
            Annotated[Item, "Some metadata"],
            dict[str, Item],
        ],
    )
    def test_iter_dataclass_type_collections(self, collection_type):
        """Test dataclass iteration with collection types."""

        @dataclass
        class Container:
            items: collection_type  # type: ignore

        result = list(iter_dataclass_type(Container))
        assert len(result) == 2
        assert Container in result
        assert Item in result

    def test_iter_dataclass_type_duplicates(self):
        """Test that duplicate types are handled correctly."""

        @dataclass
        class Shared:
            value: int

        @dataclass
        class Container:
            first: Shared
            second: Shared
            items: list[Shared]

        result = list(iter_dataclass_type(Container))
        # Should only get each type once, despite multiple references
        assert len(result) == 2
        assert Container in result
        assert Shared in result
        # Verify Shared appears only once in the result
        assert result.count(Shared) == 1

    def test_iter_dataclass_type_self_reference(self):
        """Test handling of self-referential dataclasses."""

        @dataclass
        class TreeNode:
            value: int
            children: list["TreeNode"]

        result = list(iter_dataclass_type(TreeNode))
        # Should only get TreeNode once despite self-reference
        assert len(result) == 1
        assert TreeNode in result

    def test_iter_dataclass_type_mixed_types(self):
        """Test complex scenario with mixed type annotations."""

        @dataclass
        class Inner:
            id: int

        @dataclass
        class Middle:
            inner: Inner
            option: Inner | None
            items: list[Inner]

        @dataclass
        class Outer:
            middle: Middle
            direct: Inner
            mapping: dict[str, Middle]

        result = list(iter_dataclass_type(Outer))
        # Should get Outer, Middle, Inner (each only once)
        assert len(result) == 3
        assert Outer in result
        assert Middle in result
        assert Inner in result


class TestMergeDicts:
    def test_merge_simple_dicts(self):
        """Test merging two simple dictionaries without conflicts."""
        a = {"x": 1, "y": 2}
        b = {"z": 3}
        result = merge_dicts(a, b)
        expected = {"x": 1, "y": 2, "z": 3}
        assert result == expected

    def test_merge_nested_dicts(self):
        """Test merging nested dictionaries without conflicts."""
        a = {"database": {"host": "localhost", "port": 5432}}
        b = {"database": {"name": "mydb"}, "cache": {"enabled": True}}
        result = merge_dicts(a, b)
        expected = {
            "database": {"host": "localhost", "port": 5432, "name": "mydb"},
            "cache": {"enabled": True},
        }
        assert result == expected

    def test_merge_with_conflict_raises_exception(self):
        """Test that conflicting values raise an exception."""
        a = {"key": "value1"}
        b = {"key": "value2"}
        with pytest.raises(Exception, match="Conflict at key"):
            merge_dicts(a, b)

    def test_merge_nested_conflict_raises_exception(self):
        """Test that nested conflicting values raise an exception with correct path."""
        a = {"level1": {"level2": {"key": "old_value"}}}
        b = {"level1": {"level2": {"key": "new_value"}}}
        with pytest.raises(Exception, match="Conflict at level1.level2.key"):
            merge_dicts(a, b)

    def test_merge_empty_dicts(self):
        """Test merging with empty dictionaries."""
        a: dict = {}
        b = {"key": "value"}
        result = merge_dicts(a, b)
        assert result == {"key": "value"}

        a = {"key": "value"}
        b = {}
        result = merge_dicts(a, b)
        assert result == {"key": "value"}

    def test_merge_identical_dicts(self):
        """Test merging dictionaries with identical values."""
        a = {"key": "same_value"}
        b = {"key": "same_value"}
        result = merge_dicts(a, b)
        assert result == {"key": "same_value"}

    def test_merge_modifies_first_dict_in_place(self):
        """Test that the first dictionary is modified in place."""
        a = {"original": "value"}
        b = {"new": "data"}
        result = merge_dicts(a, b)
        assert a is result  # Should be the same object
        assert "new" in a  # Should have modified the original

    def test_merge_with_priority_a(self):
        a = {"key": "value1"}
        b = {"key": "value2"}
        result = merge_dicts(a, b, priority="a")
        assert result == {"key": "value1"}

    def test_merge_with_priority_b(self):
        a = {"key": "value1"}
        b = {"key": "value2"}
        result = merge_dicts(a, b, priority="b")
        assert result == {"key": "value2"}



from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any
import pytest
from eyconf.config import ConfigExtra
from eyconf.config.extra_fields import AccessProxy


@dataclass
class Config42:
    int_field: int = 42
    str_field: str = "FortyTwo!"


@pytest.fixture
def conf42() -> ConfigExtra[Config42]:
    return ConfigExtra(Config42())


class TestCreation:
    def test_init(self):
        config = ConfigExtra(Config42())

        assert isinstance(config.data, AccessProxy)
        assert config.data.int_field == 42
        assert config.data.str_field == "FortyTwo!"

    def test_init_dict(self):
        config = ConfigExtra({"int_field": 10, "str_field": "Ten"}, schema=Config42)

        assert isinstance(config.data, AccessProxy)
        assert config.data.int_field == 10
        assert config.data.str_field == "Ten"

        # Should raise if not schema provided
        with pytest.raises(ValueError):
            ConfigExtra({"int_field": 10, "str_field": "Ten"})


class TestDataProperties:
    def test_schema_data(self, conf42: ConfigExtra[Config42]):
        schema_data = conf42.schema_data
        assert isinstance(schema_data, Config42)
        assert schema_data.int_field == 42
        assert schema_data.str_field == "FortyTwo!"

    def test_extra_data(self, conf42: ConfigExtra[Config42]):
        conf42.proxy.new_field = "New Value"
        assert isinstance(conf42.proxy._extra_data, dict)
        assert conf42.proxy._extra_data["new_field"] == "New Value"
        assert conf42.proxy["new_field"] == "New Value"
        assert conf42.proxy.new_field == "New Value"


class TestUpdate:
    def test_unknown_field(self):
        config = ConfigExtra(Config42())

        config.update({"unknown_field": "I am unknown!"})

        assert config.proxy.unknown_field == "I am unknown!"
        assert config._extra_data["unknown_field"] == "I am unknown!"
        with pytest.raises(AttributeError):
            _ = config._data.non_existent_field  # type: ignore[attr-defined]

    def test_dict(self):
        @dataclass
        class SchemaDict:
            folders: dict[str, Config42] = field(
                default_factory=lambda: {"placeholder": Config42()}
            )

        config = ConfigExtra(SchemaDict())

        assert config.data.folders["placeholder"].int_field == 42
        for folder in config.data.folders.values():
            assert isinstance(folder, Config42)

        config.update(
            {
                "folders": {
                    "config1": {"int_field": 1, "str_field": "One"},
                    "config2": {"int_field": 2, "str_field": "Two"},
                }
            }
        )

        assert config.data.folders["config1"].int_field == 1
        assert config.data.folders["config1"].str_field == "One"
        assert config.data.folders["config2"].int_field == 2
        assert config.data.folders["config2"].str_field == "Two"

        # We should revise this with a merge strategy later
        with pytest.raises(KeyError):
            config.data.folders["placeholder"]

    def test_dict_nested(self):
        @dataclass
        class NestedSchema:
            folders: dict[str, Config42] = field(
                default_factory=lambda: {"placeholder": Config42()}
            )

        @dataclass
        class SchemaWithNested:
            nested: NestedSchema = field(default_factory=NestedSchema)

        config = ConfigExtra(SchemaWithNested())

        assert config.data.nested.folders["placeholder"].int_field == 42
        for folder in config.data.nested.folders.values():
            assert isinstance(folder, Config42)

        config.update(
            {
                "nested": {
                    "folders": {
                        "config1": {"int_field": 1, "str_field": "One"},
                        "config2": {"int_field": 2, "str_field": "Two"},
                    },
                    "unknown_field": "I am unknown!",
                }
            }
        )

        assert config.data.nested.folders["config1"].int_field == 1
        assert config.data.nested.folders["config1"].str_field == "One"
        assert config.data.nested.folders["config2"].int_field == 2
        assert config.data.nested.folders["config2"].str_field == "Two"
        assert config.proxy.nested.unknown_field == "I am unknown!"

    @pytest.mark.skip(reason="Changed design, no longer using attribute dicts")
    def test_unknown_nested(self):
        config = ConfigExtra(Config42())

        config.update({"level_one": {"level_two": {"level_three": "Deep Value"}}})
        assert config.proxy.level_one.level_two.level_three == "Deep Value"


class TestToDict:
    """Tests for the to_dict method.
    This should resolve AccessProxy instances correctly.
    """

    def test_to_dict_no_extra(self, conf42: ConfigExtra[Config42]):
        result = conf42.to_dict()
        expected = {
            "int_field": 42,
            "str_field": "FortyTwo!",
        }
        assert result == expected

    def test_to_dict_with_extra(self, conf42: ConfigExtra[Config42]):
        conf42.proxy.new_field = "New Value"
        result = conf42.to_dict()
        expected = {
            "int_field": 42,
            "str_field": "FortyTwo!",
            "new_field": "New Value",
        }
        assert result == expected

        # Should also allow extra_fields=False
        result_no_extra = conf42.to_dict(extra_fields=False)
        expected_no_extra = {
            "int_field": 42,
            "str_field": "FortyTwo!",
        }
        assert result_no_extra == expected_no_extra


class TestAccessProxy:
    @pytest.fixture
    def proxy(self):
        config_data = Config42()
        extra_data: dict[str, Any] = dict()
        return AccessProxy(
            config_data,
            extra_data,
        )

    def test_attribute_assignment(self, proxy):
        proxy.int_field = 100
        proxy.new_field = "baz"

        assert proxy.int_field == 100
        assert proxy.new_field == "baz"
        assert proxy._data.int_field == 100
        assert proxy._extra_data["new_field"] == "baz"

    @pytest.mark.skip(reason="Changed design, no longer using attribute dicts")
    def test_attribute_assignment_nested(self, proxy):
        proxy.nested.level = 42
        assert proxy.nested.level == 42
        assert proxy._extra_data.nested.level == 42

    def test_attribute_assignment_other_ap(self, proxy):
        """
        Access Proxies should never hold other Access Proxies, this can only happen
        if we (accidentally) set a custom field manually.
        """
        proxy.nested = AccessProxy(
            Config42(),
            {"lorem": "ipsum"},
        )

        assert not isinstance(proxy.nested, AccessProxy)

        unpacked = proxy._to_dict()
        assert unpacked == {
            "int_field": 42,
            "str_field": "FortyTwo!",
            "nested": {
                "int_field": 42,
                "str_field": "FortyTwo!",
                "lorem": "ipsum",
            },
        }

    def test_item_assignment(self, proxy):
        proxy["int_field"] = 100
        proxy["new_field"] = "baz"

        assert proxy["int_field"] == 100
        assert proxy["new_field"] == "baz"

    def test_item_assignment_nested(self, proxy):
        proxy["nested"] = {}
        proxy["nested"]["level"] = 42

        assert proxy["nested"]["level"] == 42
        assert proxy._extra_data["nested"]["level"] == 42

    def test_to_dict(self):
        config_data = Config42()
        extra_data: dict[str, Any] = dict()
        proxy = AccessProxy(
            config_data,
            extra_data,
        )
        proxy.foo = "bar"

        result = proxy._to_dict()
        expected = {
            "int_field": 42,
            "str_field": "FortyTwo!",
            "foo": "bar",
        }
        assert result == expected

    def test_to_dict_nested(self):
        @dataclass
        class NestedConfig:
            nested_42: Config42

        config_data = NestedConfig(nested_42=Config42())
        extra_data: dict[str, Any] = dict()
        proxy = AccessProxy(
            config_data,
            extra_data,
        )

        result = proxy._to_dict()
        expected = {
            "nested_42": {
                "int_field": 42,
                "str_field": "FortyTwo!",
            }
        }
        assert result == expected

    def test_deepcopy_equality(self):
        config_data = Config42()
        extra_data: dict[str, Any] = dict()
        proxy = AccessProxy(
            config_data,
            extra_data,
        )

        proxy_2 = deepcopy(proxy)
        assert proxy_2 == proxy
        assert proxy_2._to_dict() == proxy._to_dict()

    def test_hash(self):
        config_data = Config42()
        extra_data: dict[str, Any] = dict(list_for_hash=[1, 2, 3])
        proxy = AccessProxy(
            config_data,
            extra_data,
        )

        proxy_2 = deepcopy(proxy)
        assert hash(proxy_2) == hash(proxy)


class TestReset:
    def test_simple(self, conf42: ConfigExtra[Config42]):
        conf42.update(
            {
                "int_field": 100,
                "str_field": "Updated value!",
            }
        )

        assert conf42.data.int_field == 100
        assert conf42.data.str_field == "Updated value!"

        conf42.reset()

        assert conf42.data.int_field == 42
        assert conf42.data.str_field == "FortyTwo!"



from __future__ import annotations

from dataclasses import asdict, dataclass, field

from typing import Any, TypeVar
import pytest
from eyconf import Config, ConfigExtra
from eyconf.asdict import asdict_with_aliases
from eyconf.decorators import allow_additional, dict_access
from eyconf.validation import (
    MultiConfigurationError,
    validate,
)

T = TypeVar("T", bound=Any)


@dataclass
class Config42:
    int_field: int = 42
    str_field: str = "FortyTwo!"


@dataclass
class ConfigNested:
    nested: Config42 = field(default_factory=Config42)
    nested_optional: Config42 | None = None
    other_field: str = "Hello, World!"


@pytest.fixture
def conf42() -> Config[Config42]:
    return Config(Config42(), schema=Config42)


@pytest.fixture
def conf_nested() -> Config[ConfigNested]:
    return Config(ConfigNested(), schema=ConfigNested)


@dataclass
class AliasConfig:
    attr_field: int = field(metadata={"alias": "dict_field"})
    str_field: str = "FortyTwo!"


@allow_additional
@dataclass
class AliasConfigAdditional:
    attr_field: int = field(metadata={"alias": "dict_field"})
    str_field: str = "FortyTwo!"

    # __allow_additional: ClassVar[bool] = True


@dataclass
class NestedAliasConfig:
    nested: AliasConfig = field(default_factory=lambda: AliasConfig(attr_field=43))
    other_field: str = "Hello, World!"


@dict_access
@dataclass
class AliasDictConfig:
    attr_field: int = field(metadata={"alias": "dict_field"})
    str_field: str = "FortyTwo!"


class TestAlias:
    def test_utils(self):
        config = AliasConfig(attr_field=42)
        dump = asdict_with_aliases(config)
        assert dump["dict_field"] == 42

        nested_config = NestedAliasConfig()
        nested_dump = asdict_with_aliases(nested_config)
        assert nested_dump["nested"]["dict_field"] == 43

    def test_validate(self):
        config = AliasConfig(attr_field=42)

        # this should not raise, since we have aliases
        validate(config, schema=AliasConfig)
        validate(asdict_with_aliases(config), schema=AliasConfig)

        with pytest.raises(MultiConfigurationError):
            # using the dataclasses native asdict
            # will not do our alias mapping so it should fail
            validate(asdict(config), schema=AliasConfig)

    def test_validate_additional(self):
        config = AliasConfig(attr_field=42)

        config.foo = "bar"  # type: ignore

        with pytest.raises(MultiConfigurationError):
            # By default, no additional attributes are allowed.
            validate(config, schema=AliasConfig)

        validate(config, schema=AliasConfigAdditional)

    def test_dict_alias_update(self):
        config = Config(AliasConfig(attr_field=42))

        # Update via alias
        # update also calls validate in the background
        config.update({"dict_field": 16})

        assert config.data.attr_field == 16

    def test_constructor_alias(self):
        # Constructor with dict using alias
        config = Config({"dict_field": 100, "str_field": "Test"}, schema=AliasConfig)

        assert config.data.attr_field == 100
        assert config.data.str_field == "Test"

    def test_extra_data_nested_aliased(self):
        @dataclass
        class ConfigAliasedParent:
            import_: Config42 = field(
                default_factory=lambda: Config42(), metadata={"alias": "import"}
            )

        config = ConfigExtra(ConfigAliasedParent())
        assert config.data.import_.int_field == 42
        assert config.proxy["import"].int_field == 42

        config.proxy.import_.new_field = "New Value"
        assert config.proxy.import_.new_field == "New Value"
        assert config.proxy._extra_data["import"]["new_field"] == "New Value"
        assert config._extra_data["import"]["new_field"] == "New Value"
        assert config.proxy["import"].new_field == "New Value"
        assert config.proxy["import"]["new_field"] == "New Value"


class TestToDictAlias:
    """Test that aliases are resolved properly when using the to_dict function."""

    @pytest.mark.parametrize(
        "config_class",
        [
            Config,
            ConfigExtra,
        ],
    )
    def test_to_dict_with_aliases(
        self,
        config_class: type[Config | ConfigExtra],
    ):
        config = config_class(
            data=AliasConfig(attr_field=55, str_field="AliasTest"), schema=AliasConfig
        )

        dump = config.to_dict()
        assert dump["dict_field"] == 55
        assert dump["str_field"] == "AliasTest"
        assert "attr_field" not in dump



"""Configuration base classes for EYconf.

Allows to create configuration classes based on dataclass schemas,
with validation and serialization capabilities.
"""

from __future__ import annotations

import logging
from copy import deepcopy
from dataclasses import asdict, is_dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeVar,
    cast,
)

from eyconf.asdict import asdict_with_aliases
from eyconf.generate_yaml import dataclass_to_yaml
from eyconf.type_utils import (
    get_type_hints_resolve_namespace,
    is_dataclass_instance,
    is_dataclass_type,
)
from eyconf.utils import dataclass_from_dict, dict_items_resolve_aliases
from eyconf.validation import to_json_schema, validate, validate_json

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

# Needs the string escaping to work at runtime as _typeshed is not a real module
D = TypeVar("D", bound="DataclassInstance")

log = logging.getLogger(__name__)


class Config(Generic[D]):
    """Base configuration class.

    This class allows to create configuration objects based on dataclass schemas.
    It provides methods for validation, updating, overwriting, and converting
    configuration data.

    This can be used to create custom configuration classes in memory without file I/O.
    """

    _schema: type[D]
    _data: D
    _json_schema: dict

    def __init__(
        self,
        data: dict | D,
        schema: type[D] | None = None,
    ):
        if is_dataclass_type(data):
            raise ValueError("Data must be a dict or datacalss instance, not schema!")

        if schema is not None:
            self._schema = schema
        else:
            if not is_dataclass(data):
                raise ValueError(
                    "If no schema is provided, data must be of the schema dataclass instance."
                )
            self._schema = type(data)

        # Create schema, raise if Schema is invalid
        self._json_schema = to_json_schema(self._schema)

        # Will raise ConfigurationError if the data does not comply with the schema
        validate(data, self._json_schema)

        if is_dataclass(data):
            self._data = cast(D, data)
        else:
            self._data = dataclass_from_dict(self._schema, data)

    def validate(self):
        """Validate the current data against the schema."""
        validate(self._data, self._json_schema)

    def update(self, data: dict[str, Any]):
        """Update the configuration with provided data.

        This applies an partial update to the existing configuration data.
        Only the provided keys will be updated, others will remain unchanged.
        """

        def _update(
            target_type: type[DataclassInstance],
            target: DataclassInstance,
            update_data: dict[str, Any],
            path: list[tuple[DataclassInstance, str]] | None = None,
        ):
            """Recursive update helper function.

            Parameters
            ----------
            target_type : type[DataclassInstance]
                The dataclass type of the target instance.
            target : DataclassInstance
                The dataclass instance to update.
            update_data : dict
                The update data to apply.
            path : list[tuple[DataclassInstance, str]] | None
                The current path in the dataclass tree (root-to-leaf).
                [0] is parent instance,
                [1] is attr key used to access the target from from its parent. E.g.
                [(root_instance, "child_field"), (child_instance, "grandchild_field")]
            """
            target_annotations = get_type_hints_resolve_namespace(target_type)

            for key, value in dict_items_resolve_aliases(update_data, target_type):
                if hasattr(target, key):
                    current_value = getattr(target, key)

                    # Handle dataclass fields
                    if is_dataclass_instance(current_value):
                        _update(
                            target_annotations[key],
                            current_value,
                            value,
                            path=(path or []) + [(target, key)],
                        )

                    # Handle Optional fields that were previously None
                    elif current_value is None and isinstance(value, dict):
                        nested_instance = dataclass_from_dict(
                            target_annotations[key], value
                        )
                        setattr(target, key, nested_instance)
                    elif current_annotation := target_annotations.get(key):
                        nested = dataclass_from_dict(
                            current_annotation,
                            # TODO: If we want to implement a merge strategy
                            # some like this could work:
                            # merge_dicts(asdict(current_value), value),  # type: ignore[arg-type]
                            value,
                        )
                        setattr(target, key, nested)
                    else:
                        # Primitives and direct assignments
                        # Can only be reached if a dynamic field is added
                        # to the dataclass instance
                        setattr(target, key, value)
                else:
                    # Non-schema fields
                    self._update_additional(
                        value,
                        path=(path or []) + [(target, key)],
                    )

        old_data = deepcopy(self._data)
        _update(self._schema, self._data, data)

        try:
            self.validate()
        except Exception:
            self._data = old_data
            raise

    def _update_additional(
        self,
        value: Any,
        path: list[tuple[DataclassInstance, str]],
    ) -> None:
        """Handle updating additional (non-schema) fields.

        Can be overwritten in subclasses.
        """
        target, attr_key = path[-1]
        root_name = type(path[0][0]).__name__
        path_str = f"{root_name}" + "".join(f".{attr}" for _, attr in path)
        raise AttributeError(
            f"Cannot set non-schema field '{attr_key}' "
            f"on dataclass '{type(target).__name__}'!"
            f"\n(at {path_str})"
        )

    def overwrite(self, data: dict | D):
        """Overwrite the configuration with provided data.

        If the provided data is missing required fields, an error will be raised.
        """
        data = asdict(data) if is_dataclass(data) else data
        validate_json(data, self._json_schema)
        self._data = dataclass_from_dict(self._schema, data)

    def reset(self):
        """Reset the configuration data to the default values."""
        self._data = self._schema()

    # -------------------------------- Converters -------------------------------- #

    @property
    def data(self) -> D:
        """Get the configuration data."""
        return self._data

    def to_dict(self) -> dict:
        """Convert the configuration data to a dictionary."""
        return asdict_with_aliases(self._data)

    def to_yaml(self) -> str:
        """Convert the configuration data to a yaml string."""
        return dataclass_to_yaml(self._data)

    # --------------------------------- Printing --------------------------------- #

    def __repr__(self) -> str:
        """Return a custom string representation of the configuration object."""
        class_name = type(self).__name__
        memory_address = hex(id(self))
        prefix = f"<{class_name} object at {memory_address}>:\n"
        return f"{prefix}{self.__str__()}"

    def __str__(self):
        """Return a custom formatted string representation of the configuration data."""
        data_dict = asdict(self._data)
        s = "\n".join(
            "  " + line for line in self._pretty_format(data_dict).splitlines()
        )
        return s

    def _pretty_format(self, data, indent=0):
        """Format the dict with pretty indentation."""
        result: list[str] = []
        for key, value in data.items():
            if isinstance(value, dict):
                result.append(" " * indent + f"{key}:")
                result.append(self._pretty_format(value, indent + 4))
            else:
                result.append(" " * indent + f"{key}: {value}")
        return "\n".join(result)



"""Optional CLI interface for eyconf configuration management.

Allows to easily integrate configuration management into command line applications.

Usage:
------
```python
from eyconf.cli import config_cli

app = typer.Typer()
app.subcommand(config_cli, name="config")
```
"""

import asyncio
import os

import typer

from eyconf import EYConf


def create_config_cli(
    Config: type[EYConf],
    *args,
    **kwargs,
):
    """Create a CLI for managing the configuration file.

    Parameter
    ---------
    config : type[EYConf]
        The EYConf class to manage the configuration for. Not the instance!
    *args, **kwargs
        Additional arguments and keyword arguments to pass to the
        Config class if it needs to be instantiated.

    Usage
    -----
    ```python
    from eyconf.cli import create_config_cli
    from my_config import MyConfig

    config_cli = create_config_cli(EYConf, schema=MyConfig)

    app = typer.Typer()
    app.add_typer(config_cli, name="config")
    ```
    """
    config_cli = typer.Typer(
        rich_markup_mode="rich",
        help="Manage configuration file",
    )

    @config_cli.callback(invoke_without_command=True)
    def main(
        ctx: typer.Context,
        edit: bool = typer.Option(
            False, "--edit", "-e", help="Edit the configuration."
        ),
    ):
        """Edit the plistsync configuration."""
        if edit:
            asyncio.run(edit_config(Config, *args, **kwargs))
        else:
            # Show help if no subcommand is provided
            if ctx.invoked_subcommand is None:
                print(ctx.get_help())

    @config_cli.command()
    def ls():
        """Show the current configuration."""
        config = Config(*args, **kwargs)
        typer.echo(str(config))

    return config_cli


async def edit_config(Config: type[EYConf], *args, **kwargs):
    """Edit the configuration file."""
    path = Config.get_file()

    if not path.exists():
        # If the config file does not exist, create it with default values
        Config(*args, **kwargs)

    # Open the config file with the default system editor
    process = None
    typer.echo(f"Opening configuration file: {path.absolute().as_posix()}")
    try:
        if os.name == "nt":  # Windows
            os.startfile(path.absolute().as_posix())  # type: ignore[attr-defined]
        elif os.name == "posix":  # macOS or Linux
            process = await asyncio.create_subprocess_exec(
                *[
                    "open" if os.uname().sysname == "Darwin" else "xdg-open",
                    path.absolute().as_posix(),
                ]
            )
        else:
            typer.echo("Unsupported OS for the edit command.")
    except Exception as e:
        typer.echo(f"Failed to open the configuration editor: {e}")

    await process.wait() if process else None



# In CI: We run this without `from __future__ import annotations` to ensure compatibility.
# from __future__ import annotations

from dataclasses import dataclass, field
from typing import Annotated, Literal, Optional

import logging
import pytest
import yaml

from eyconf.generate_yaml import dataclass_to_yaml

log = logging.getLogger(__name__)


class TestGenerateDefault:
    """Test the generation of default yaml from a dataclass."""

    def test_basic(self):
        """Test the basic functionality."""

        @dataclass
        class Basic:
            int_field: int = 42
            str_field: str = "Hello, World!"
            bool_field: bool = True
            float_field: float = 3.14

        yaml_str = dataclass_to_yaml(Basic)
        assert (
            yaml_str
            == "int_field: 42\nstr_field: Hello, World!\nbool_field: true\nfloat_field: 3.14"
        )
        assert yaml.safe_load(yaml_str) == {
            "int_field": 42,
            "str_field": "Hello, World!",
            "bool_field": True,
            "float_field": 3.14,
        }

    def test_optional(self):
        @dataclass
        class Optionals:
            int_field: int | None
            str_field: str | None
            other_field: str | None

        yaml_str = dataclass_to_yaml(Optionals)
        print(yaml_str)
        assert yaml_str == "int_field: null\nstr_field: null\nother_field: null"
        assert yaml.safe_load(yaml_str) == {
            "int_field": None,
            "str_field": None,
            "other_field": None,
        }

    def test_optional_with_default(self):
        @dataclass
        class OptionalsWithDefault:
            int_field: int | None = 42
            str_field: str | None = "test"
            other_field: str | None = None

        yaml_str = dataclass_to_yaml(OptionalsWithDefault)
        assert yaml_str == "int_field: 42\nstr_field: test\nother_field: null"
        assert yaml.safe_load(yaml_str) == {
            "int_field": 42,
            "str_field": "test",
            "other_field": None,
        }

    def test_nested(self):
        """Nested dataclasses should raise an error."""

        @dataclass
        class Nested:
            int_field: int = 42
            str_field: str = "Hello, World!"

        @dataclass
        class Parent:
            nested: Nested
            other_field: str = "Hello, World!"

        yaml_str = dataclass_to_yaml(Parent)
        print(yaml_str)
        assert (
            yaml_str
            == "nested:\n  int_field: 42\n  str_field: Hello, World!\n\nother_field: Hello, World!"
        )
        assert yaml.safe_load(yaml_str) == {
            "nested": {"int_field": 42, "str_field": "Hello, World!"},
            "other_field": "Hello, World!",
        }

    def test_deep_nested(self):
        @dataclass
        class Nested:
            int_field: int = 42
            str_field: str = "Hello, World!"

        @dataclass
        class Parent:
            nested: Nested
            other_field: str = "Hello, World!"

        @dataclass
        class GrandParent:
            parent: Parent
            other_field: str = "Hello, World!"

        yaml_str = dataclass_to_yaml(GrandParent)
        print(yaml_str)
        assert (
            yaml_str
            == "parent:\n  nested:\n    int_field: 42\n    str_field: Hello, World!\n\n  other_field: Hello, World!\n\nother_field: Hello, World!"
        )
        assert yaml.safe_load(yaml_str) == {
            "parent": {
                "nested": {"int_field": 42, "str_field": "Hello, World!"},
                "other_field": "Hello, World!",
            },
            "other_field": "Hello, World!",
        }

    def test_list(self):
        """Test a list of dataclasses."""

        @dataclass
        class Lists:
            int_list: list[int] | None
            str_list: list[str] = field(default_factory=lambda: ["a", "b", "c"])

        yaml_str = dataclass_to_yaml(Lists)
        print(yaml_str)
        assert yaml_str == "int_list: null\nstr_list:\n  - a\n  - b\n  - c"
        assert yaml.safe_load(yaml_str) == {
            "int_list": None,
            "str_list": ["a", "b", "c"],
        }

    def test_dict(self):
        """Test a dict of dataclasses."""

        @dataclass
        class DictSchema:
            empty: dict[str, str]
            int_dict: dict[str, int] = field(
                default_factory=lambda: {"one": 1, "two": 2, "three": 3}
            )
            str_dict: dict[str, str] | None = None

        yaml_str = dataclass_to_yaml(DictSchema)
        print(yaml_str)
        assert yaml_str == (
            "empty: {}\nint_dict:\n  one: 1\n  two: 2\n  three: 3\nstr_dict: null"
        )

        with pytest.raises(
            ValueError, match="Only string keys are supported in dict types"
        ):

            @dataclass
            class InvalidDictSchema:
                invalid_dict: dict[int, str] = field(
                    default_factory=lambda: {1: "one", 2: "two"}
                )

            yaml_str = dataclass_to_yaml(InvalidDictSchema)

    def test_list_union(self):
        @dataclass
        class Lists:
            int_list: list[int] | int = field(default_factory=lambda: [1, 2, 3])
            str_list: list[str] | str = "test"

        yaml_str = dataclass_to_yaml(Lists)
        print(yaml_str)
        assert yaml_str == "int_list:\n  - 1\n  - 2\n  - 3\nstr_list: test"

    def test_complex(self):
        """Test a more complex dataclass."""

        @dataclass
        class Movie:
            title: str = "Jurassic Park"
            year: int = 1993

        @dataclass
        class Director:
            name: str = "Steven Spielberg"
            movies: list[Movie] = field(
                default_factory=lambda: [
                    Movie(title="Jurassic Park", year=1993),
                    Movie(title="Indiana Jones", year=1981),
                ]
            )

        yaml_str = dataclass_to_yaml(Director)
        yaml_parsed = yaml.safe_load(yaml_str)

        assert yaml_parsed == {
            "name": "Steven Spielberg",
            "movies": [
                {"title": "Jurassic Park", "year": 1993},
                {"title": "Indiana Jones", "year": 1981},
            ],
        }

    def test_recursive(self):
        """Test a recursive dataclass."""

        @dataclass
        class Recursive:
            next: Optional["Recursive"] = None

        yaml_str = dataclass_to_yaml(Recursive)
        print(yaml_str)
        assert yaml_str == "next: null"
        assert yaml.safe_load(yaml_str) == {"next": None}

    def test_derived(self):
        """Test a dataclass with derived fields."""

        @dataclass
        class Base:
            enabled: bool = False

        @dataclass
        class Derived(Base):
            name: str = "Test"

        yaml_str = dataclass_to_yaml(Derived)
        print(yaml_str)
        assert yaml_str == "enabled: false\nname: Test"
        assert yaml.safe_load(yaml_str) == {"enabled": False, "name": "Test"}

    def test_advanced_deep_nested(self):
        @dataclass
        class InboxFolder:
            name: str
            path: str
            auto_threshold: float | None = None
            autotag: Literal["auto", "preview", "bootleg", False] = False
            # the `no` -> False option will need tweaking

        @dataclass
        class LibrarySection:
            readonly: bool = False
            artist_separators: list[str] = field(
                default_factory=lambda: [",", ";", "&"]
            )

        @dataclass
        class TerminalSection:
            start_path: str = "/repo"

        @dataclass
        class InboxSection:
            ignore: list[str] | Literal["_use_beets_ignore"] = "_use_beets_ignore"
            debounce_before_autotag: int = 30
            folders: dict[str, InboxFolder] = field(default_factory=dict)

        @dataclass
        class BeetsFlaskSchema:
            inbox: InboxSection = field(default_factory=lambda: InboxSection())
            library: LibrarySection = field(default_factory=lambda: LibrarySection())
            terminal: TerminalSection = field(default_factory=lambda: TerminalSection())
            num_preview_workers: int = 4

        @dataclass
        class BeetsSchema:
            gui: BeetsFlaskSchema

        yaml_str = dataclass_to_yaml(BeetsSchema)
        print(yaml_str)
        assert True


class TestDocstringGeneration:
    """Test generation of docstrings from dataclass fields."""

    def test_simple(self):
        """Test the basic functionality."""

        @dataclass
        class SingleLineDocstring:
            """This is a docstring."""

            pass

        yaml_str = dataclass_to_yaml(SingleLineDocstring)
        assert yaml_str == "# This is a docstring.\n"
        assert yaml.safe_load(yaml_str) == None

    def test_multiline(self):
        @dataclass
        class MultiLineDocstring:
            """This is a docstring that is longer than 80 characters and should be split into multiple lines."""

            pass

        yaml_str = dataclass_to_yaml(MultiLineDocstring)
        assert (
            yaml_str
            == "# This is a docstring that is longer than 80 characters and should be split into\n# multiple lines.\n"
        )
        assert yaml.safe_load(yaml_str) == None

    def test_no_docstring(self):
        @dataclass
        class NoDocstring:
            pass

        yaml_str = dataclass_to_yaml(NoDocstring)
        assert yaml_str == ""
        assert yaml.safe_load(yaml_str) == None

    def test_annotation_docstring(self):
        """Test the basic functionality."""

        @dataclass
        class VariableDocstring:
            """This is a module docstring."""

            int_field: Annotated[int, "This is a docstring for the int field."] = 42

        yaml_str = dataclass_to_yaml(VariableDocstring)
        print(yaml_str)
        assert (
            yaml_str
            == "# This is a module docstring.\n\n# This is a docstring for the int field.\nint_field: 42"
        )

    def test_multiline_variable_docstring(self):
        """Test the basic functionality."""

        @dataclass
        class MultiLineVariableDocstring:
            """This is a module docstring."""

            int_field: Annotated[int, "This is a docstring", "Test"] = 42

        yaml_str = dataclass_to_yaml(MultiLineVariableDocstring)
        print(yaml_str)
        assert (
            yaml_str
            == "# This is a module docstring.\n\n# This is a docstring\n# Test\nint_field: 42"
        )
        assert yaml.safe_load(yaml_str) == {"int_field": 42}

    def test_nested_variable_docstring(self):
        """Test the basic functionality."""

        @dataclass
        class Inner:
            """This is a section docstring."""

            single: Annotated[int, "This is a docstring for the int field."] = 42
            multi: Annotated[int, "This is a docstring", "Test"] = 42

        @dataclass
        class Outer:
            """This is a module docstring."""

            inner: Annotated[Inner, "This is a docstring for the inner field."] = field(
                default_factory=lambda: Inner()
            )

        yaml_str = dataclass_to_yaml(Outer)
        expected_yaml = (
            "# This is a module docstring.\n\n"
            "# This is a docstring for the inner field.\n"
            "inner:\n"
            "  # This is a section docstring.\n\n"
            "  # This is a docstring for the int field.\n"
            "  single: 42\n"
            "  # This is a docstring\n"
            "  # Test\n"
            "  multi: 42\n"
        )
        assert yaml_str == expected_yaml
        assert yaml.safe_load(yaml_str) == {"inner": {"single": 42, "multi": 42}}



from __future__ import annotations

import logging
from copy import deepcopy
from dataclasses import is_dataclass
from functools import cache
from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast

from eyconf.asdict import asdict_with_aliases
from eyconf.type_utils import (
    is_dataclass_instance,
    is_dataclass_type,
    iter_dataclass_type,
)
from eyconf.utils import (
    Metadata,
    dataclass_from_dict,
    merge_dicts,
    metadata_fields_from_dataclass,
)
from eyconf.validation import validate
from eyconf.validation._to_json import to_json_schema

from .base import Config

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

# Needs the string escaping to work at runtime as _typeshed is not a real module
D = TypeVar("D", bound="DataclassInstance")


log = logging.getLogger(__name__)


class AccessProxy(Generic[D]):
    """Proxy to access attributes dynamically."""

    _data: D  # this should never be an access proxy!
    _extra_data: dict

    # this proxies' location in the overall config tree
    _parent: AccessProxy | None

    def __init__(
        self,
        data: D,
        extra_data: dict,
        parent: AccessProxy | None = None,
    ):
        self._data = data
        self._extra_data = extra_data
        self._parent = parent

    @property
    @cache
    def _fields_metadata(self) -> dict[str, Metadata]:
        """Get the fields of the current dataclass schema."""
        return metadata_fields_from_dataclass(self._data)

    def _resolve_attr_to_dict_key(self, attr_key: str) -> str:
        """Resolve an attribute key to its dict key using aliasing."""
        return self._fields_metadata.get(attr_key, {}).get("alias", attr_key)

    def _resolve_dict_to_attr_key(self, dict_key: str) -> str:
        """Resolve a dict key to its attribute key using aliasing."""
        for attr_key, metadata in self._fields_metadata.items():
            if metadata.get("alias") == dict_key:
                return attr_key
        return dict_key

    def _to_dict(self) -> dict:
        """Convert the AccessProxy to a standard dictionary."""
        extra = deepcopy(self._extra_data)
        data = deepcopy(asdict_with_aliases(self._data))
        result = merge_dicts(data, extra)
        return result

    def __getattr__(self, attr_key: str) -> Any:
        """Get field via attribute style access (non-aliased keys)."""
        dict_key: str = self._resolve_attr_to_dict_key(attr_key)
        try:
            data = getattr(self._data, attr_key)
            if is_dataclass_instance(data):
                if dict_key not in self._extra_data:
                    self._extra_data[dict_key] = dict()
                return AccessProxy(
                    data=data,
                    extra_data=self._extra_data[dict_key],
                    parent=self,
                )
            else:
                return data
        except AttributeError:
            return self._extra_data[dict_key]

    def __setattr__(self, attr_key: str, value: Any):
        """Set field via attribute style access (non-aliased keys)."""
        if isinstance(value, AccessProxy):
            value = value._to_dict()

        if attr_key.startswith("_"):
            object.__setattr__(self, attr_key, value)
        else:
            if hasattr(self._data, attr_key):
                setattr(self._data, attr_key, value)
            else:
                dict_key = self._resolve_attr_to_dict_key(attr_key)
                self._extra_data[dict_key] = value

    def __getitem__(self, dict_key: str) -> Any:
        """Get field via dict style acces (alias)."""
        attr_key = self._resolve_dict_to_attr_key(dict_key)
        if hasattr(self._data, attr_key):
            return self.__getattr__(attr_key)
        else:
            return self._extra_data[dict_key]

    def __setitem__(self, dict_key: str, value: Any) -> None:
        """Set field via dict style access (alias)."""
        attr_key = self._resolve_dict_to_attr_key(dict_key)
        if hasattr(self._data, attr_key):
            self.__setattr__(attr_key, value)
        else:
            self._extra_data[dict_key] = value

    def __deepcopy__(self, memo):
        """Implement deepcopy to avoid issues with nested dataclasses."""
        new_proxy = type(self)(
            data=deepcopy(self._data, memo),
            extra_data=deepcopy(self._extra_data, memo),
            parent=self._parent,
        )
        return new_proxy

    def __hash__(self) -> int:
        """Make Access Proxy hashable by recursively hashing its internal data."""

        def make_hashable(obj):
            if isinstance(obj, dict):
                return tuple(sorted((k, make_hashable(v)) for k, v in obj.items()))
            elif isinstance(obj, list):
                return tuple(make_hashable(item) for item in obj)
            else:
                return obj

        return hash(make_hashable(self._to_dict()))

    def __eq__(self, value: object, /) -> bool:
        """Compare Access Proxies via their data."""
        if isinstance(value, AccessProxy):
            value = value._to_dict()
        return self._to_dict() == value


class ConfigExtra(Config[D]):
    """Configuration class that supports extra fields explicitly.

    This class extends the base configuration functionality to allow
    for additional fields that are not defined in the original schema.

    This additional fields are accessible via the `extra_data` property. They
    are merged into the schema data when accessing via the `data` property.
    """

    _extra_data: dict[str, Any]
    _access_proxy: AccessProxy[D]

    def __init__(
        self,
        data: dict | D,
        schema: type[D] | None = None,
    ):
        if is_dataclass_type(data):
            raise ValueError("Data must be a dict or datacalss instance, not schema!")

        if schema is not None:
            self._schema = schema
        else:
            if not is_dataclass(data):
                raise ValueError(
                    "If no schema is provided, data must be of the schema dataclass instance."
                )
            self._schema = type(data)

        # Automatically set __allow_additional to True in the schema(s) if not set
        for s in iter_dataclass_type(self._schema):
            if not hasattr(s, "__allow_additional"):
                setattr(s, "__allow_additional", True)
            else:
                log.debug(
                    f"Schema {s.__name__} already has __allow_additional set to "
                    f"{getattr(s, '__allow_additional')}."
                )

        # Create schema, raise if Schema is invalid
        self._json_schema = to_json_schema(self._schema)

        # Will raise ConfigurationError if the data does not comply with the schema
        validate(data, self._json_schema)

        self._extra_data = dict()

        if is_dataclass(data):
            self._data = cast(D, data)
        else:
            self._data = dataclass_from_dict(self._schema, data)

        self._access_proxy = AccessProxy(
            data=self._data,
            extra_data=self._extra_data,
        )

    def reset(self):
        """Reset the configuration data to the default values."""
        super().reset()
        self._access_proxy = AccessProxy(
            data=self._data,
            extra_data=self._extra_data,
        )

    @property
    def data(self) -> D:
        """Get the configuration data wrapped in a dynamic accessor.

        Care: Instance checks will not work as expected on this property.

        We have two non-ideal choices for the type hint here:
        - if we use D, we will get a wrong type_error for e.g. `config.data.my_field.to_dict()`
        - if we use AccessProxy[D], we wont get the nice type_checking against the schema.
        """
        return self._access_proxy  # type: ignore

    @property
    def proxy(self) -> AccessProxy[D]:
        """Convenience Property for setting non-schema fields."""
        return self._access_proxy

    @property
    def schema_data(self) -> D:
        """Get the schema dataclass type excluding extra fields."""
        return self._data

    @property
    def extra_data(self) -> dict:
        """Get the extra data as a dict."""
        return self._extra_data

    def to_dict(self, extra_fields: bool = True) -> dict:
        """Get the full configuration data as a dictionary, including extra fields."""
        data = asdict_with_aliases(self.proxy._data)
        if extra_fields:
            data = merge_dicts(data, self.proxy._extra_data)
        return data

    def _update_additional(
        self,
        value: Any,
        path: list[tuple[DataclassInstance, str]],
    ) -> None:
        """Handle updating additional (non-schema) fields used in `super.update`.

        Here, we update extra_data, which is a dict. So we use dict style.
        """
        dict_key_path = []
        for target, attr_key in path:
            # resolve attr_key to dict_key using aliasing
            field_metadata = metadata_fields_from_dataclass(target)
            dict_key = field_metadata.get(attr_key, {}).get("alias", attr_key)
            dict_key_path.append(dict_key)

        extra_data: dict[str, Any] = self._extra_data
        for dict_key in dict_key_path[:-1]:
            if dict_key not in extra_data.keys():
                extra_data[dict_key] = dict()
            extra_data = extra_data[dict_key]

        extra_data[dict_key_path[-1]] = value



from __future__ import annotations

import functools
from collections.abc import Callable
from dataclasses import is_dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Protocol,
    TypeVar,
    overload,
    runtime_checkable,
)

from eyconf.type_utils import is_dataclass_type
from eyconf.utils import get_metadata

if TYPE_CHECKING:
    from _typeshed import DataclassInstance


T = TypeVar("T")
D = TypeVar("D", bound="DataclassInstance")


@runtime_checkable
class DictAccess(Protocol):
    """Protocol for dict-like access."""

    def __getitem__(self, key: str) -> Any: ...  # noqa: D105


@runtime_checkable
class DictSetAccess(Protocol):
    """Protocol for dict-like access with setting."""

    def __setitem__(self, key: str, value: Any) -> None: ...  # noqa: D105


def _aliases_map(cls: DataclassInstance) -> dict[str, str]:
    """Get mapping of aliases to field names for a dataclass."""
    return {m["alias"]: f.name for f, m in get_metadata(cls) if "alias" in m}


def _get_attr_resolve_alias(self: DataclassInstance, key: str) -> Any:
    """Get item resolving aliases."""
    aliases = _aliases_map(self)
    if key in aliases.keys():
        return getattr(self, aliases[key])
    elif key in aliases.values():
        _suggestion = next((k for k, v in aliases.items() if v == key), None)
        raise KeyError(
            "If an alias is defined, subscripting is only allowed "
            + f"using the alias. Use ['{_suggestion}'] instead of ['{key}']!"
        )

    return getattr(self, key)


def _set_attr_resolve_alias(self: DataclassInstance, key: str, value: Any) -> None:
    """Set item resolving aliases."""
    aliases = _aliases_map(self)
    if key in aliases.keys():
        return setattr(self, aliases[key], value)
    elif key in aliases.values():
        _suggestion = next((k for k, v in aliases.items() if v == key), None)
        raise KeyError(
            "If an alias is defined, subscripting is only allowed "
            + f"using the alias. Use ['{_suggestion}'] instead of ['{key}']!"
        )

    return setattr(self, key, value)


@overload
def dict_access(cls: type[T]) -> type[T]: ...
@overload
def dict_access(
    cls: None = None, getter: bool = True, setter: bool = False
) -> Callable[[type[T]], type[T]]: ...


def dict_access(
    cls: type[T] | None = None, getter: bool = True, setter: bool = False
) -> type[T] | Callable[[type[T]], type[T]]:
    """Class decorator to add dict-like access to class attributes.

    Can be used to add `dict`-like access to any class, allowing
    attribute access via the `obj['attribute']` syntax.

    Use with care, dict-style access does not provide any type safety
    and will not be checked by static type checkers.

    Usage:

    ```python
    @dict_access
    class MySchema:
        forty_two: int = 42

    #1 (cls) -> cls


    @dict_access(setter=True)
    class MySchema:

    #2 (args) -> (cls) -> cls

    obj = MySchema()
    assert isinstance(obj, DictAccess)
    print(obj['forty_two'])  # Outputs: 42
    ```
    """

    def _decorate(cls: type[T]) -> type[T]:
        @functools.wraps(cls)
        def wrap(target_cls: type[T]) -> type[T]:
            if getter:
                setattr(target_cls, "__getitem__", _get_attr_resolve_alias)
            if setter:
                setattr(target_cls, "__setitem__", _set_attr_resolve_alias)
            return target_cls

        return wrap(cls)

    if cls is None:
        return _decorate
    else:
        return _decorate(cls)


def allow_additional(cls: type[T]) -> type[T]:
    """Class decorator to allow additional attributes in dataclass.

    This prevent validation errors if a dataclass instance holds an attribute
    that is not defined in the dataclass schema.

    Note this will not automatically load additional (none-schema) fields from the yaml.
    To also load additional fields use `EyConfExtraFields` as base class for your
    configuration dataclass.
    """
    mangled = f"_{cls.__name__}__allow_additional"
    setattr(cls, mangled, True)
    return cls


def check_allows_additional(schema: D | type[D]) -> bool:
    """Whether the dataclass allows additional properties."""
    if is_dataclass_type(schema):
        return getattr(schema, f"_{schema.__name__}__allow_additional", False)
    elif is_dataclass(schema) and not isinstance(schema, type):
        return getattr(schema, f"_{schema.__class__.__name__}__allow_additional", False)
    return False



from __future__ import annotations

import inspect
import logging
import sys
from collections.abc import Iterator, Sequence
from dataclasses import is_dataclass
from types import UnionType
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    TypeGuard,
    TypeVar,
    get_args,
    get_origin,
    get_type_hints,
)

# for some reason typing  Sequence and abc sequence are not the same type
from typing import Sequence as TypingSequence  # noqa: UP035

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

D = TypeVar("D", bound="DataclassInstance")

log = logging.getLogger(__name__)


def get_type_hints_resolve_namespace(obj, include_extras: bool = False):
    """Get type hints for an object, resolving namespaces for dataclasses.

    Workaround for when using `from __future__ import annotations`.

    This is needed because Types become strings, and native
    recursive resolution does not work on strings alone.
    To resolve types recursively in this case, the Dataclasses are needed
    and can be passed to `get_type_hints` via `globalns` and `localns`.
    """
    try:
        return get_type_hints(obj, include_extras=include_extras)
    except NameError:
        globalns, localns = _get_namespace(obj)
        return get_type_hints(
            obj,
            globalns=globalns,
            localns=localns,
            include_extras=include_extras,
        )


def _get_namespace(
    obj,
) -> tuple[dict[str, Any], dict[str, Any]]:
    """Get the global and local namespaces for a dataclass.

    I.e. where to find the dataclass definitions and possibly
    other types used in the dataclass.

    Workaround for when using `from __future__ import annotations`.

    This is needed because Types become strings, and native
    recursive resolution does not work on strings alone.
    To resolve types recursively in this case, the Dataclasses are needed
    and can be passed to `get_type_hints` via `globalns` and `localns`.
    """
    module = sys.modules.get(obj.__module__)
    globalns = getattr(module, "__dict__", {}) if module else {}

    # Get the caller's locals to handle locally defined classes
    frame = inspect.currentframe()
    localns = {}

    # Walk up the call stack to find local variables
    while frame:
        frame_locals = frame.f_locals

        # Merge locals that might contain our dataclass definitions
        for name, value in frame_locals.items():
            if inspect.isclass(value):
                localns[name] = value
                globalns[name] = value  # Also add to globals

        frame = frame.f_back

    return globalns, localns


def is_dataclass_type(obj: Any) -> TypeGuard[type]:
    """Check if an object is a dataclass type (class), not an instance."""
    return is_dataclass(obj) and isinstance(obj, type)


def is_dataclass_instance(obj: Any) -> TypeGuard[DataclassInstance]:
    """Check if an object is a dataclass instance."""
    return is_dataclass(obj) and not isinstance(obj, type)


def iter_dataclass_type(schema: type[D]) -> Iterator[type[DataclassInstance]]:
    """Iterate over all dataclass nested instances in the given dataclass type.

    Duplicate types are automatically handled by using a set to track visited types.

    Yields
    ------
    DataclassInstance
        Each nested dataclass instance found within the schema (also the root).
    """
    visited = set()
    stack = [schema]

    def _add_type_to_stack(*t: type[Any]) -> None:
        """Add type to stack if it is a dataclass and not yet visited."""
        for item in t:
            if is_dataclass_type(item) and id(item) not in visited:
                stack.append(item)

    while stack:
        current_type = stack.pop()
        type_id = id(current_type)
        # Skip if we've already visited this type
        if type_id in visited:
            continue

        visited.add(type_id)
        yield current_type

        # Process fields of the current dataclass
        type_hints = get_type_hints_resolve_namespace(current_type, include_extras=True)
        for _, field_type in type_hints.items():
            origin = get_origin(field_type)

            if origin is Annotated:
                # Unpack Annotated types
                field_type = get_args(field_type)[0]
                origin = get_origin(field_type)

            if origin in {UnionType, list, tuple, set, Sequence, TypingSequence, dict}:
                # Handle collection types
                _add_type_to_stack(*get_args(field_type))

            if is_dataclass_type(field_type):
                _add_type_to_stack(field_type)



"""Contains mostly a copy of dataclasses.asdict, but with support for field aliases."""

from __future__ import annotations

import copy
import logging
import types
from dataclasses import fields
from typing import (
    TYPE_CHECKING,
    TypeVar,
)

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

log = logging.getLogger(__name__)

D = TypeVar("D", bound="DataclassInstance")


def asdict_with_aliases(
    obj: DataclassInstance | DataclassInstance,
    include_attributes: bool = True,
    include_properties: bool = False,
) -> dict:
    """
    Convert a dataclass to a dict, applying aliases if present.

    Every attribute in the dataclass that has an `alias` defined in the metadata
    will use the alias as the key in the resulting dictionary (instead of the attribute name).

    This is pretty much a copy of the standard `dataclasses.asdict` function,
    but we had to modify it, because the dict_factory needs to access obj recursively.
    (previously dataclass obj not passed down)
    """
    if not hasattr(type(obj), "__dataclass_fields__"):
        raise TypeError("asdict_with_aliases() should be called on dataclass instances")

    return _asdict_inner(
        obj,
        dict_factory=_alias_dict_factory,
        include_attributes=include_attributes,
        include_properties=include_properties,
    )  # type: ignore[arg-type]


# -------------- Slightly modified copy from dataclasses.asdict -------------- #

_ATOMIC_TYPES = frozenset(
    {
        # Common JSON Serializable types
        types.NoneType,
        bool,
        int,
        float,
        str,
        # Other common types
        complex,
        bytes,
        # Other types that are also unaffected by deepcopy
        types.EllipsisType,
        types.NotImplementedType,
        types.CodeType,
        types.BuiltinFunctionType,
        types.FunctionType,
        type,
        range,
        property,
    }
)


def _asdict_inner(obj, **kwargs):
    dict_factory = kwargs.get("dict_factory", dict)
    include_attributes = kwargs.get("include_attributes", False)
    include_properties = kwargs.get("include_properties", False)

    if type(obj) in _ATOMIC_TYPES:
        return obj
    elif hasattr(type(obj), "__dataclass_fields__"):
        # obj is dataclass
        # fast path for the common case
        if dict_factory is dict:
            result = {
                f.name: _asdict_inner(getattr(obj, f.name), dict_factory=dict)
                for f in fields(obj)
            }
        else:
            _result = []
            for f in fields(obj):
                value = _asdict_inner(getattr(obj, f.name), **kwargs)
                _result.append((f.name, value))
            result = dict_factory(obj, _result)

        field_names = {f.name for f in fields(obj)}
        extra_attrs = {}
        if include_attributes:
            for k, v in obj.__dict__.items():
                if not k.startswith("_") and k not in field_names:
                    extra_attrs[k] = _asdict_inner(v, **kwargs)

        if include_properties:
            for prop in dir(obj):
                if (
                    not prop.startswith("_")
                    and prop not in field_names
                    and prop not in extra_attrs
                ):
                    attr = getattr(type(obj), prop, None)
                    if isinstance(attr, property):
                        try:
                            extra_attrs[prop] = _asdict_inner(
                                getattr(obj, prop), **kwargs
                            )
                        except Exception:
                            pass  # Ignore property errors

        # Merge. Can we be sure that result is a dict? Depends on factory.
        if isinstance(result, dict):
            result.update(extra_attrs)

        return result

    elif isinstance(obj, tuple) and hasattr(obj, "_fields"):
        # obj is a namedtuple.  Recurse into it, but the returned
        # object is another namedtuple of the same type.  This is
        # similar to how other list- or tuple-derived classes are
        # treated (see below), but we just need to create them
        # differently because a namedtuple's __init__ needs to be
        # called differently (see bpo-34363).

        # I'm not using namedtuple's _asdict()
        # method, because:
        # - it does not recurse in to the namedtuple fields and
        #   convert them to dicts (using dict_factory).
        # - I don't actually want to return a dict here.  The main
        #   use case here is json.dumps, and it handles converting
        #   namedtuples to lists.  Admittedly we're losing some
        #   information here when we produce a json list instead of a
        #   dict.  Note that if we returned dicts here instead of
        #   namedtuples, we could no longer call asdict() on a data
        #   structure where a namedtuple was used as a dict key.

        return type(obj)(*[_asdict_inner(v, dict_factory=dict_factory) for v in obj])
    elif isinstance(obj, (list, tuple)):
        # Assume we can create an object of this type by passing in a
        # generator (which is not true for namedtuples, handled
        # above).
        return type(obj)(_asdict_inner(v, dict_factory=dict_factory) for v in obj)
    elif isinstance(obj, dict):
        if hasattr(type(obj), "default_factory"):
            # obj is a defaultdict, which has a different constructor from
            # dict as it requires the default_factory as its first arg.
            result = type(obj)(getattr(obj, "default_factory"))  # type: ignore
            for k, v in obj.items():
                result[_asdict_inner(k, dict_factory=dict_factory)] = _asdict_inner(
                    v, dict_factory=dict_factory
                )
            return result
        return type(obj)(
            (
                _asdict_inner(k, dict_factory=dict_factory),
                _asdict_inner(v, dict_factory=dict_factory),
            )
            for k, v in obj.items()
        )
    else:
        return copy.deepcopy(obj)


def _alias_dict_factory(obj, items):
    result = {}
    for key, value in items:
        # Find the field object for this key
        field_obj = next((f for f in fields(obj) if f.name == key), None)
        if field_obj and "alias" in field_obj.metadata:
            result[field_obj.metadata["alias"]] = value
        else:
            result[key] = value
    return result
